;       AX.25 rx routine                        N0QBH   2000
;       4/01/2000
;       10 Mhz XTAL
;       No digi paths displayed
;       VISUAL 1 on after 2 flags rx'd
;       TNC style serial data output
;       MAX storage     53 bytes
;____________________________________________________________
	LIST	P=PIC16F84
	include	"P16F84.INC"
	__CONFIG _HS_OSC & _CP_OFF & _PWRTE_ON & _WDT_OFF
;____________________________________________________________
;               Port Set-up
;           0   1   2   3   4   5   6   7       bit
; port  A   Out Out Out In  In  X   X   X
;       B   Out Out Out In  In  Out Out Out
;____________________________________________________________
	ORG	0			; start code
	goto	main
;____________________________________________________________
        cblock  0x0c
                aux_count               ;flag counter
                byte_count              ;delay counter
                store_count             ;bytes stored cnt
                serial_count            ;serial out counter
                bit_stuff_count         ;counter
                bit_count               ;counter
                send_reg                ;serial data
                FLAGS                   ;various flags
                delay_count             ;delay timer
                byte_in                 ;data byte rx'd
                AX25_crc_hi             ;upper byte CRC word
                AX25_crc_lo             ;lower byte CRC word
                next_to_last_byte       ;2nd from last byte
                last_byte               ;last byte rxed
        endc
;------------------------------------------------------------
#define SERIAL_DATA_OUT      PORTA,1
#define PTT_OUT              PORTA,2
#define RX_FILTER            PORTA,4
#define VISUAL_ONE           PORTB,0
#define MODE1_OUT            PORTB,1
#define MODE0_OUT            PORTB,2
#define BIT_IN               PORTB,4
#define DATA_FLAG            FLAGS,1
#define SHIFTED_DATA         FLAGS,2
;------------------------------------------------------------
intro_message
        addwf   PCL,F
        dt      " PIC-E packet decoder *"
        dt      " N0QBH 2000"
        dt      0x0d,0x0a,0x00
;------------------------------------------------------------
;------------------------------------------------------------
;
main    
        bsf     STATUS,RP0              ;bank 1
        movlw   b'11111000'             ;A:3,:4 inputs
        movwf   TRISA                   ;set up port A I/O
        movlw   b'00011000'             ;B:3,4 input
        movwf   TRISB                   ;set up port B I/O
        movlw   b'11010011'
        movwf   OPTION_REG              ;TMR0 /16
        bcf     STATUS,RP0              ;bank 0
        clrf    INTCON                  ;clear INT reg
        bcf     PTT_OUT                 ;unkey radio
        bsf     MODE0_OUT               ;MX614 setup
        bsf     MODE1_OUT               ;lo pwr mode
        bcf     SERIAL_DATA_OUT         ;stop bit set
        movlw   0x0a                    ;10 loops
        call    adjustable_delay        ;2 mS delay
        bcf     MODE0_OUT               ;RX mode
        movlw   0x64                    ;100 loops
        call    adjustable_delay        ;20 mS delay
        call    startup_message         ;sign on msg
;------------------------------------------------------------
house_clean                             ;new packet
        clrf    aux_count               ;reset misc counter
        clrf    delay_count             ;reset timer
        clrf    byte_in                 ;clear data byte
        clrf    bit_stuff_count         ;reset b_s_count
        clrf    store_count             ;reset count out
        clrf    byte_count              ;reset byte count
        clrf    FLAGS                   ;reset flags
        movlw   0xff
        movwf   AX25_crc_hi             ;reset CRC calc
        movwf   AX25_crc_lo
        movlw   0x1a
        movwf   FSR                     ;reset FSR
        movlw   0x7e
        movwf   TMR0
        movfw   PORTB                   ;reset RBI
        clrf    INTCON                  ;set up TMR0, RBI
        bcf     VISUAL_ONE              ;clear flags lite
find_start_flag
        clrc
        call    recv_one_bit            ;check for toggle
        skpnz
        setc                            ;no toggle (1)
        rrf     byte_in,F               ;toggle  (0)
        movfw   byte_in                 ;byte_in to W
        xorlw   0x7e                    ;is it a flag ?
        skpz
        goto    find_start_flag         ;no
start_flag_loop                         ;yes, look again
        call    recv_one_bit
        skpnz                           ;see a toggle ?
        goto    find_start_flag         ;no, go back
        movlw   0x06                    ;yes, must see 6 1's
        movwf   store_count
long_ones
        call    recv_one_bit            ;look for 6
        skpz                            ;1's in arow
        goto    find_start_flag         ;saw a 0
        decfsz  store_count,F           ;saw a 1
        goto    long_ones
        call    recv_one_bit            ;look for end 0
        skpnz
        goto    find_start_flag         ;not flag, reset
        bsf     VISUAL_ONE              ;flag, LED on
;------------------------------------------------------------
;
get_next_byte
        clrf    byte_in                 ;clear input byte
        movlw   0x08
        movwf   bit_count               ;get 8 bits
get_next_bit
        call    recv_one_bit
        skpz
        goto    zero_in                 ; 0 rx'd
        incf    bit_stuff_count,F       ;inc counter
        setc                            ; 1 rx'd
roll_em_in
        rrf     byte_in,F               ;bit stored
        movfw   bit_stuff_count
        xorlw   0x05                    ;5 1's in a row ?
        skpnz
        goto    check_stuff             ;yes
count_em_out                            ;no
        decfsz  bit_count,F             ;8 bits ?
        goto    get_next_bit            ;no
        goto    CRC_save_store          ;yes
zero_in
        clrf    bit_stuff_count         ;0 seen, clear bsc
        clrc
        goto    roll_em_in
;------------------------------------------------------------
; Bit stuff decode routine
;
check_stuff
        call    recv_one_bit            ;wait 1 bit
        skpnz                           ;if data=0, skip next
        goto    sixth_one               ;unchanged, flag ?
        clrf    bit_stuff_count         ;tog seen, clear count
        goto    count_em_out            ;don't count this bit
sixth_one
        call    recv_one_bit            ;get 6th bit
        skpnz                           ;flag ?
        goto    house_clean             ;no, 7th 1, reset
        movfw   bit_count
        xorlw   0x03
        skpz                            ;in the right place ?
        goto    house_clean             ;no
        clrf    bit_stuff_count         ;yes, flag, clear bstc
        btfss   DATA_FLAG               ;header/tail flag ?
        goto    get_next_byte           ;header flag
        movfw   byte_count              ;tail flag
        sublw   0x0f                    ;how long is packet ?
        skpnc
        goto    house_clean             ; < 17 bytes rx'd
        goto    packet_end              ;it's long enough
;
CRC_save_store
        btfss   DATA_FLAG               ;test data flag
        bsf     DATA_FLAG               ;set data flag
        incf    byte_count,W            ;count bytes
        sublw   0xf0                    ;how long ?
        skpc
        goto    house_clean             ; > 240 bytes, reset
        incf    byte_count,F
        movfw   byte_count              ;how long again
        sublw   0x02
        skpc                            ; >2 ?
        call    AX25_CRC_calc           ;yes
        movfw   last_byte               ;no, last byte to W
        movwf   next_to_last_byte       ;to 2nd to last
        movfw   byte_in                 ;get new data byte
        movwf   last_byte               ;store in last byte
        goto    get_next_byte           ;get next byte
;
;------------------------------------------------------------
;       Invert CRC data and compare to value sent
;
packet_end
        movlw   0xff
        xorwf   AX25_crc_lo,F
        xorwf   AX25_crc_hi,F           ;invert  CRC data
        movfw   next_to_last_byte       ;check lobyte match
        xorwf   AX25_crc_lo,W
        skpz
        goto    house_clean             ;mismatch, reset
        movfw   last_byte
        xorwf   AX25_crc_hi,W           ;check hibyte match
        skpz
        goto    house_clean             ;mismatch, reset
        movfw   FSR                     ;FSR last byte
        movwf   store_count             ;loc stored
        decf    store_count,F           ;store_count - 1
;------------------------------------------------------------
;       Serial send routine begins here
reset_FSR_pointer
        movlw   0x21                    ;org call start
        movwf   FSR                     ;in memory
        call    parse_call              ;send callsign
        movlw   0x27                    ;org call SSID
        movwf   FSR                     ;location
        call    parse_ssid              ;send SSID if any
send_r_arrow
        movlw   '>'                     ;insert r_arrow
        movwf   send_reg
        call    serial_out              ;send >
        movlw   0x1a                    ;start of dest call
        movwf   FSR
        call    parse_call              ;send callsign
        movlw   0x20
        movwf   FSR
        call    parse_ssid              ;send SSID if any
        movlw   ':'
        movwf   send_reg
        call    serial_out              ;send :
        movlw   0x28
        movwf   FSR                     ;start of info
info_field_loop
        call    read_buffer
        call    serial_out              ;send byte
        incf    FSR,F                   ;inc FSR
        movfw   FSR                     ;FSR to W
        subwf   store_count,W           ;compare to end
        skpnc
        goto    info_field_loop         ;<= store_count
        movlw   0x0d                    ;>, send end
        movwf   send_reg                ;<CR>
        call    serial_out
        movlw   0x0a                    ;<LF>
        movwf   send_reg
        call    serial_out
        goto    house_clean
;------------------------------------------------------------
;       Subroutines area
;------------------------------------------------------------
startup_message
        clrf    aux_count               ;offset clear
startup_loop
        movfw   aux_count               ;offset to W
        call    intro_message           ;get offset byte
        movwf   send_reg                ;store
        xorlw   0x00                    ;test for end
        skpnz
        return                          ;done, start prgm
        call    serial_out              ;send byte
        incf    aux_count,F             ;inc offset
        goto    startup_loop            ;loop
;------------------------------------------------------------
AX25_CRC_calc
        movfw   byte_count              ;where are we ?
        sublw   0x10
        skpnc                           ; < 17 bytes ?
        goto    call_field_store        ;yes
        btfsc   SHIFTED_DATA            ;no, past the addr ?
        goto    info_field_store        ;yes
        call    e_o_a_test              ;no, test for end
        goto    crc_rotor               ;calc crc
info_field_store
        btfss   aux_count,1             ;counted 2 ?
        goto    two_byte_skip           ;no
        movfw   FSR                     ;yes
        sublw   0x4f                    ;buffer full?
        skpc
        goto    crc_rotor               ;yes, skip store
        movfw   next_to_last_byte       ;straight store
        movwf   INDF                    ;store at offset
        incf    FSR,F                   ;inc FSR
        goto    crc_rotor
two_byte_skip
        incf    aux_count,F             ;aux_count + 1
        goto    crc_rotor
call_field_store
        call    e_o_a_test              ;test for end
        clrc
        rrf     next_to_last_byte,W     ;stores data in
        movwf   INDF                    ;readable format
        incf    FSR,F                   ;inc FSR
crc_rotor
        movlw   0x08
        movwf   bit_count               ; 8 loops
crc_head
        rrf     next_to_last_byte,F     ; get bit for CRC
        call    crc_core                ;do CRC calc
        decfsz  bit_count,F             ;8 counts
        goto    crc_head
        return
;------------------------------------------------------------
crc_core
        movlw   1
        skpnc
        xorwf   AX25_crc_lo, F          ; crc calculation
        clrc
        rrf     AX25_crc_hi, F
        rrf     AX25_crc_lo, F
        skpc
        return
        movlw   H'08'
        xorwf   AX25_crc_lo, F
        movlw   H'84'
        xorwf   AX25_crc_hi, F
        return
;------------------------------------------------------------
e_o_a_test
        movfw   next_to_last_byte       ;test for end
        andlw   0x01                    ;of address flag
        skpnz
        return                          ;not seen
        bsf     SHIFTED_DATA            ;1st set LSB sets
        clrf    aux_count               ;reset 2 byte skip
        return                          ;saw e_o_a flag
;------------------------------------------------------------
read_buffer                             ;get FSR pointer
        movfw   INDF                    ;data and put in
        movwf   send_reg                ;send_reg
        return
;------------------------------------------------------------
;       Assumes FSR is set
parse_call
        movlw   0x06
        movwf   aux_count               ;6 loops
parse_call_loop
        call    read_buffer             ;get byte from INDF
        movfw   send_reg                ;send_reg to W
        xorlw   ' '                     ;space ?
        skpnz
        return                          ;yes, done
        call    serial_out              ;no, send byte
        incf    FSR,F                   ;FSR + 1
        decfsz  aux_count,F             ;loop count - 1
        goto    parse_call_loop
        return                          ;saw space or 6 loops
;------------------------------------------------------------
;       assumes FSR is set to SSID loc
parse_ssid
        movlw   0x0f
        andwf   INDF,F                  ;mask hi nibble
        call    read_buffer             ;read data
        movfw   send_reg                ;send_reg to W
        skpnz                           ;SSID = 0  ?
        return                          ;yes, done
        movlw   '-'                     ;no, add dash
        movwf   send_reg
        call    serial_out              ;send -
        call    read_buffer             ;read again
        movlw   0x0a                    ;get SSID magnitude
        subwf   send_reg,W
        skpc
        goto    ssid_ones               ;<10, send  digit
        movlw   0x31                    ;>9, ascii 1 to W
        movwf   send_reg                ;W to send_reg
        call    serial_out              ;send 10
        call    read_buffer             ;read INDF again
        movlw   0x0a
        subwf   send_reg,F              ;send_reg -10
ssid_ones
        movlw   0x30                    ;
        addwf   send_reg,F              ;convert to ascii
        call    serial_out              ;send ones byte
        return
;------------------------------------------------------------
;------------------------------------------------------------
;       Packet bit receive routine
;       Returns with 0 or 1 in Zero Flag
recv_one_bit                            ;delay 833 uS
        btfsc   INTCON,RBIF
        goto    bit_edge                ;saw tone toggle
        btfss   INTCON,T0IF
        goto    recv_one_bit            ;not timed out
        movlw   0x7e                    ;timed out
        movwf   TMR0                    ;reset timer
        bcf     INTCON,T0IF             ;1 byte long
        clrw                            ;Z flag set
        return
bit_edge
        movlw   0x3d                    ;load 1 1/2 bits
        movwf   TMR0
        movfw   PORTB                   ;reset port change
        bcf     INTCON,RBIF             ;clear RBIF
        bcf     INTCON,T0IF
        movfw   TMR0                    ;Z flag clear
        return
;------------------------------------------------------------
;               4800,8,N,1      serial out routine
;
serial_out
        movlw   0x08                    ;init shift counter
        movwf   serial_count
        bsf     SERIAL_DATA_OUT         ;start bit set
        call    onebit_delay
        comf    send_reg,F              ;invert data
        clrc
serial_out_loop
        rrf     send_reg,F              ;rotate bit 0 into carry
        skpnc                           ;test carry flag
        goto    hbit                    ;carry = 1
        bcf     SERIAL_DATA_OUT         ;bit is clear, output 0
        goto    time2_roll
hbit    bsf     SERIAL_DATA_OUT         ;bit is set, output 1
time2_roll
        call    onebit_delay
        decfsz  serial_count,F          ;shifted 8?
        goto    serial_out_loop         ;no
        bcf     SERIAL_DATA_OUT         ;yes, output stopbit
        call    onebit_delay
        return                          ;done
;------------------------------------------------------------
;                10 Mhz version
onebit_delay                            ;208 uS delay (4k8)
        movlw   0xad                    ;173 loops @ 1.2 uS
        movwf   delay_count
onebit_loop
        decfsz  delay_count,F
        goto    onebit_loop
        return
;------------------------------------------------------------
adjustable_delay                        ;W contains length
        movwf   aux_count
the_mS_loop
        call    onebit_delay
        decfsz  aux_count,F
        goto    the_mS_loop
        return
;------------------------------------------------------------
        END
